#!/usr/bin/env python
# -*- coding: latin-1 -*-
###############################################################################
##
##      ##  ######
##       ######  ###
##  ## ###############        Shark 3D (www.shark3d.com)
##   ########## # # #
##    ########                Copyright (c) 1996-2020 Spinor GmbH.
##   ######### # # #          All rights reserved.
##  ##   ##########
##      ##
##
###############################################################################

"""
Copy all needed files to vhd file .
"""

import re
import os
import subprocess
import shutil
import contextlib
import winreg
import uuid

###############################################################################

def createHiddenStartUpInfo():
    startupInfo = subprocess.STARTUPINFO()
    startupInfo.dwFlags = subprocess.STARTF_USESHOWWINDOW
    startupInfo.wShowWindow = subprocess.SW_HIDE
    return startupInfo

###############################################################################

def copy_folder(src, dest):
    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        items = os.listdir(src)
        for item in items:
            copy_from_path = os.path.join(src, item).replace("/", "\\")
            copy_to_path = os.path.join(dest, item).replace("/", "\\")
            copy_folder(copy_from_path, copy_to_path)
    else:
        if not os.path.exists(os.path.dirname(dest)):
            os.makedirs(os.path.dirname(dest))
            copy_folder(src, dest)
        else:
            shutil.copy(src, dest)

###############################################################################
                    
@contextlib.contextmanager
def temporarilyRegSubKey(key, subKey):
    try:
        handle = winreg.OpenKey(key, subKey, 0, winreg.KEY_ALL_ACCESS)
    except FileNotFoundError:
        try:
            sep = subKey.rindex("\\")
        except ValueError:
            handle = winreg.CreateKey(key, subKey)
            try:
                yield handle
            finally:
                winreg.CloseKey(handle)
                winreg.DeleteKey(key, subKey)
        else:
            parentKey = subKey[:sep]
            with temporarilyRegSubKey(key, parentKey):
                handle = winreg.CreateKey(key, subKey)
                try:
                    yield handle
                finally:
                    winreg.CloseKey(handle)
                    winreg.DeleteKey(key, subKey)
    else:
        try:
            yield handle
        finally:
            winreg.CloseKey(handle)
    
@contextlib.contextmanager
def temporarilyRegValue(key, valName, ty, val):
    try:
        oldVal, oldTy = winreg.QueryValueEx(key, valName)
    except FileNotFoundError:
        try:
            winreg.SetValueEx(key, valName, None, ty, val)
            yield
        finally:
            winreg.DeleteValue(key, valName)
    else:
        try:
            winreg.SetValueEx(key, valName, None, ty, val)
            yield
        finally:
            winreg.SetValueEx(key, valName, None, oldTy, oldVal)

@contextlib.contextmanager
def temporarilyRegSubKeyValue(key, subKey, valName, ty, val):
    with temporarilyRegSubKey(key, subKey) as handle:
        with temporarilyRegValue(handle, valName, ty, val):
            yield

@contextlib.contextmanager
def temporarilyDisableAutoRun():
    with temporarilyRegSubKeyValue(
            winreg.HKEY_CURRENT_USER, 
            "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Policies\\Explorer",
            "NoDriveTypeAutoRun", winreg.REG_DWORD, 0xff):
        yield
        
###############################################################################

@contextlib.contextmanager
def mountVhd(vhdPath):
    vhdPathNormalized = vhdPath.replace("/", "\\")
    with temporarilyDisableAutoRun():
        print("Mounting VHD")
        p = subprocess.Popen(
                ["diskpart"], 
                stdin=subprocess.PIPE, 
                stdout=subprocess.PIPE, 
                stderr=subprocess.PIPE, 
                startupinfo=createHiddenStartUpInfo()
            )
        stdout_data, stderr_data = p.communicate(
            ("select vdisk file=\"" + vhdPathNormalized + "\"\n"
             "attach vdisk\n"
             "select partition 1\n"
             "detail partition\n"
             "exit\n").encode("utf-8"))
        stdout_text = stdout_data.decode("utf-8")#
        stderr_text = stderr_data.decode("utf-8")
        
        drive = None
        for m in re.finditer(
                "\\*\\s+Volume\\s+(?P<volume>\\d+)\\s+(?P<drive>\\S+)\\s+"
                "NTFS\\s+Partition\\s+(?P<size>\\S+\\s+\\S+)\\s+Healthy", 
                stdout_text):
            drive = m.group("drive")
        if not drive:
            raise ValueError("Mounting failed.")            
        mountpath = "%s:/" % drive
        
        yield mountpath
        
        print("Unmounting VHD")
        p = subprocess.Popen(
                ["diskpart"], 
                stdin=subprocess.PIPE, 
                stdout=subprocess.PIPE, 
                stderr=subprocess.PIPE, 
                startupinfo=createHiddenStartUpInfo()
            )
        stdout_data, stderr_data = p.communicate(
            ("select vdisk file=\"" + vhdPathNormalized + "\"\n"
             "detach vdisk\n"
             "exit\n").encode("utf-8"))
        print("Copy finished.")
    
###############################################################################

class _FileSystemWorker:
    def __init__(self, basePath):
        self._basePath = basePath
        
    #Copy file or folder-content to target dir
    def copy_into(self, srcPath, dest_path):
        full_dest_path = os.path.join(self._basePath, dest_path)

        if os.path.isdir(srcPath):
            print("copy folder: ", srcPath)
            copy_folder(srcPath, full_dest_path)
        elif os.path.isfile(srcPath):
            print("copy file: ", srcPath)
            copy_folder(srcPath, full_dest_path)          
        else:
            raise ValueError("source path %r not found", srcPath)
    
    def delete_folder(self, folderPath):
        full_folder_path = os.path.join(self._basePath, folderPath)
        if os.path.exists(full_folder_path):
            shutil.rmtree(full_folder_path) 
        
    @contextlib.contextmanager
    def get_real_path_of_virtual_path(self, path):
        yield os.path.join(self._basePath, path)
        
    @contextlib.contextmanager
    def open(self, path, mode):
        full_path = os.path.join(self._basePath, path)
        yield open(full_path, mode)
    
@contextlib.contextmanager
def openVhd(vhdPath):
    with mountVhd(vhdPath) as mountPath:
        worker = _FileSystemWorker(mountPath)
        yield worker
        
###############################################################################

@contextlib.contextmanager
def loadRegistryHive(hivePath):
    hiveName = str(uuid.uuid1())
    hiveRegistryPath = "HKLM\\" + hiveName
    subprocess.call([
            "reg.exe", 
            "load", 
            hiveRegistryPath, 
            hivePath.replace("/", "\\")], 
            startupinfo=createHiddenStartUpInfo())
    try:
        print("hive loaded")
        yield hiveName
    finally:
        subprocess.call([
                "reg.exe", 
                "unload", 
                hiveRegistryPath], 
                startupinfo=createHiddenStartUpInfo())

@contextlib.contextmanager
def loadRegistryHiveOpen(hivePath):
    with loadRegistryHive(hivePath) as hiveName:
        with winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE, 
                hiveName, 
                0, 
                winreg.KEY_ALL_ACCESS) as hiveRegKey:
            yield hiveRegKey

###############################################################################

if __name__ == '__main__':    
    # if you don't have a .vhd file to copy you can download one from 
    # https://www.microsoft.com/en-in/evalcenter/evaluate-windows-server
    baseDir="xxxxxxxx/"
    vhdInput = "xxxxxxxx/WindowsServer2012withUpdates.vhd"
    vhdOutput = "xxxxxxxx/shark3dServerImage.vhd"

    if not os.path.exists(vhdOutput):
        shutil.copyfile(vhdInput, vhdOutput)

    with openVhd(vhdOutput) as vhd:
        # Delete folder shark3d on vhd if exists.
        vhd.delete_folder("shark3d/")
        vhd.delete_folder("project/")
        vhd.delete_folder("apps/")

        # Copy shark3d system files.
        shark3dSrcFilesAndFolders = [
                ("external/win/openssl", ""),
                ("external/win/vc_2013_runtime/x64/msvcr120.dll", ""),
                ("external/win/vc_2017_runtime/x64/vcruntime140.dll", ""),
                ("external/win/windows_resource_kits_tools/srvany.exe", ""),
                ("external/python3/Lib", ""),
                ("external/python3/x64/DLLs", ""),
                ("external/python3/x64/python37.dll", ""),
                ("xxxxxxxx_private_key.key", 
                        "shark3d/utility/service/vm/"),
                ("xxxxxxxx_ssl_certificate.cer", 
                        "shark3d/utility/service/vm/"),
                ("xxxxxxxx/vm_virtualwall", ""),
                # if not found then you must execute .bat file in 
                # trunk\src\app\player\
                ("bin/win_res/shark3d_player.dat", ""),
                # if not found then you must build Release lib in Visual Studio
                ("bin/win_vc2017/x64_Release_lib/shark3d_lib_webrtc_sep.dll",
                         ""),
                ("bin/win_vc2017/x64_Release_lib/shark3d_player_static.exe",
                         ""),
            ]
        for rel_src_path, dest_path in shark3dSrcFilesAndFolders:
            srcPath = os.path.join(baseDir, rel_src_path)
            if not dest_path:
                dest_path = os.path.join("shark3d", rel_src_path)
            vhd.copy_into(srcPath, dest_path)

        # Copy shark3d project pack file.
        # if win.s3d_pack was not found then you must pack it in the editor
        vhd.copy_into("xxxxxxxx/win.s3d_pack",
                 "project/project/")
        
        # Copy filebeat and certificates.
        filebeatdir = os.environ["SHARK3D_FILEBEAT_6X5X4_64BIT"]
        vhd.copy_into(filebeatdir, "apps/filebeat")
        vhd.copy_into("C:/certificates/xxxxxxxx.pem",
                "certificates/")
        vhd.copy_into("C:/certificates/xxxxxxxxkey.pem",
                "certificates/")
        
        # Create registry entry for shark3d autostart. 
        with vhd.get_real_path_of_virtual_path(
                "Windows/System32/config/SYSTEM") as vhdRegFilePath:
            with loadRegistryHiveOpen(vhdRegFilePath) as hiveRegKey:
                winreg.CreateKey(
                        hiveRegKey, "ControlSet001\\Services\\shark3d")
                # Writing registry entry into backup file of CurrentControlSet
                # which will be loaded on system startup.
                with winreg.OpenKey(
                        hiveRegKey, 
                        "ControlSet001\\Services\\shark3d", 
                        0, 
                        winreg.KEY_ALL_ACCESS) as runRegKey:
                    print("Writing registry keys.")
                    winreg.SetValueEx(
                            runRegKey, 
                            "Type", 
                            None, 
                            winreg.REG_DWORD, 
                            0x00000010)
                    winreg.SetValueEx(
                            runRegKey, 
                            "Start", 
                            None, 
                            winreg.REG_DWORD, 
                            0x00000002)
                    winreg.SetValueEx(
                            runRegKey, 
                            "ErrorControl", 
                            None, 
                            winreg.REG_DWORD, 
                            0x00000001)
                    winreg.SetValueEx(
                            runRegKey, 
                            "ImagePath", 
                            None, 
                            winreg.REG_EXPAND_SZ, 
                            "c:\\shark3d\\external\win\\" \
                                "windows_resource_kits_tools\\srvany.exe")
                    winreg.SetValueEx(
                            runRegKey, 
                            "DisplayName", 
                            None, 
                            winreg.REG_SZ, 
                            "shark3dAutoStartService")
                    winreg.SetValueEx(
                            runRegKey, 
                            "WOW64", 
                            None, 
                            winreg.REG_DWORD, 
                            0x00000001)
                    winreg.SetValueEx(
                            runRegKey, 
                            "ObjectName", 
                            None, 
                            winreg.REG_SZ, 
                            "LocalSystem")
                     
                winreg.CreateKey(
                        hiveRegKey,
                        "ControlSet001\\Services\\shark3d\\Parameters")
                with winreg.OpenKey(
                        hiveRegKey, 
                        "ControlSet001\\Services\\shark3d\\Parameters", 
                        0, 
                        winreg.KEY_ALL_ACCESS) as runRegKey:
                     
                    winreg.SetValueEx(
                            runRegKey, 
                            "Application", 
                            None, 
                            winreg.REG_SZ, 
                            "C:\\Windows\\System32\\cmd.exe")
                    
                    # Launch exe with additional parameters
                    winreg.SetValueEx(
                            runRegKey, 
                            "AppDirectory", 
                            None, 
                            winreg.REG_SZ, 
                            "c:\\")
                            
                    winreg.SetValueEx(
                             runRegKey, 
                             "AppParameters", 
                             None, 
                             winreg.REG_SZ, 
                             "/C C:\\shark3d\\utility\\service_canonical\\" \
                                    "autostart.bat")
 
 
###############################################################################
